from abc import ABC
from collections.abc import Sequence
from typing import Any, List, Optional, Callable
from langchain_core.documents import Document

from ..commons import Splitter
from ...documents.commons import TextDocument


class SplitterPostProcessor(Splitter, ABC):
    """Unified interface for post-processing chunks (e.g. recombination, re-enumerate)."""

    pass


class ReEnumeratePostProcessor(SplitterPostProcessor):
    """Post-Processor that sets a metadata field as an increasing integer, useful to track chunk sequentiality."""

    def __init__(self, index_name: str, reset_fields: Optional[List[str]] = None):
        """
        Args:
            index_name (str): The name of the metadata field to modify.
            reset_fields (Optional[List[str]], optional): The name of the fields on which to reset the counting to 0.
                Defaults to None.
        """
        self.index_name = index_name
        self.reset_fields = reset_fields or []

    def transform_documents(
        self, documents: Sequence[Document], **kwargs: Any
    ) -> Sequence[Document]:
        curr_index = 0
        curr_fields = tuple()
        for doc in documents:
            doc_fields = tuple(doc.metadata.get(field) for field in self.reset_fields)
            if doc_fields != curr_fields:
                curr_index = 0
            doc.metadata[self.index_name] = curr_index
            curr_index += 1
        return documents


class TrimEmptyPostProcessor(SplitterPostProcessor):
    """Post-Processor that filters out any document with empty text contents."""

    def __init__(self, ignore_chars: Optional[List[str]] = None):
        """

        Args:
            ignore_chars (Optional[List[str]], optional): Characters to consider as whitespace for the trimming. Defaults to None.
        """
        self.ignore_chars = ignore_chars

    def transform_documents(
        self, documents: Sequence[Document], **kwargs: Any
    ) -> Sequence[Document]:
        trimmed_documents = []
        for doc in documents:
            doc.page_content.strip(self.ignore_chars)
            if doc.page_content:
                trimmed_documents.append(doc)
        return trimmed_documents


class ReCombinePostProcessor(SplitterPostProcessor):
    def __init__(
        self,
        count_fn: Callable[[str], int],
        agg_metadata_fn: Callable[[dict, dict], dict],
        min_size: int,
        max_size: int,
        reset_fields: Optional[List[str]] = None,
    ):
        self.count_fn = count_fn
        self.agg_metadata_fn = agg_metadata_fn
        self.min_size = min_size
        self.max_size = max_size
        self.reset_fields = reset_fields or []

    def transform_documents(
        self, documents: Sequence[TextDocument], **kwargs: Any
    ) -> Sequence[TextDocument]:
        raise NotImplementedError()



from typing import Any, Dict, Optional, Union, List
from langchain_core.documents import Document
import json
import builtins
import io


class JsonDocument(Document):
    json_content: Dict[str, Any]

    def __init__(
        self,
        json_content: Union[Dict[str, Any], Any],
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        """
        A document class designed for JSON data.

        Args:
            json_content (Union[Dict[str, Any], Any]): A valid JSON dictionary.
                If another type of data is provided, it will be converted best-effort.
            metadata (Optional[Dict[str, Any]]): Metadata associated with the binary document.
        """
        if not isinstance(json_content, dict):
            try:
                json_content = json.loads(json_content)
            except:
                json_content = {"content": json_content}
        json_content = JsonDocument._safe_convert_json_keys(json_content)
        json_content = json.loads(
            json.dumps(json_content, default=TextDocument.safe_convert_to_string)
        )
        super().__init__(
            page_content="",
            json_content=json_content,
            metadata=metadata or {},
            **kwargs,
        )

    def __str__(self) -> str:
        if self.metadata:
            return f"json_content='{self.json_content}' metadata={self.metadata}"
        else:
            return f"json_content='{self.json_content}'"

    def _safe_convert_json_keys(source: dict) -> dict:
        for key, value in source.items():
            if not isinstance(key, str):
                source[str(key)] = source.pop(key)
                key = str(key)
            if isinstance(value, dict):
                source[key] = JsonDocument._safe_convert_json_keys(value)
        return source


class TextDocument(Document):
    def __init__(
        self,
        page_content: Union[str, Any],
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        """
        A document class designed for JSON data.

        Args:
            page_content (Dict[str, Any]): The text contents. If not a string, it will be converted best-effort.
            metadata (Optional[Dict[str, Any]]): Metadata associated with the binary document.
        """
        page_content = TextDocument.safe_convert_to_string(page_content)
        super().__init__(page_content=page_content, metadata=metadata or {}, **kwargs)

    @staticmethod
    def safe_convert_to_string(value: Any, skip_default: bool = False) -> Any:
        match type(value):
            case builtins.bytes:
                return value.decode("utf-8", errors="ignore")
            case _:
                if not skip_default:
                    return str(value)
        return value


class BinaryDocument(Document):
    binary_content: bytes

    def __init__(
        self,
        binary_content: Union[bytes, Any],
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        """
        A document class designed for binary data.

        Args:
            binary_content (bytes): The binary content of the document (e.g., an image, audio, or file content).
                If not bytes, the data will be converted best-effort.
            metadata (Optional[Dict[str, Any]]): Metadata associated with the binary document.
        """
        binary_content = BinaryDocument.safe_convert_to_bytes(binary_content, True)
        if not isinstance(binary_content, bytes):
            binary_content = BinaryDocument.safe_convert_to_bytes(
                TextDocument.safe_convert_to_string(binary_content)
            )
        super().__init__(
            page_content="",
            binary_content=binary_content,
            metadata=metadata or {},
            **kwargs,
        )

    def __str__(self) -> str:
        if self.metadata:
            return f"binary_content='{self.binary_content}' metadata={self.metadata}"
        else:
            return f"binary_content='{self.binary_content}'"

    @staticmethod
    def safe_convert_to_bytes(
        value: Any, skip_default: bool = False
    ) -> Union[bytes, Any]:
        if not isinstance(value, bytes):
            match type(value):
                case builtins.bytearray | builtins.memoryview:
                    value = bytes(value)
                case io.BytesIO:
                    value = value.getvalue()
                case io.FileIO:
                    value = value.read()
                case builtins.int | builtins.float:
                    value = value.tobytes()
                case _:
                    if not skip_default:
                        value = bytes(str(value), encoding="utf-8")
        return value


class VectorizedDocument(Document):
    vector_data: List[float]
    document: Document

    def __init__(self, vector_data: List[float], document: Document, **kwargs):
        """
        A wrapper around a Document, containing the embedded data for the document.

        Args:
            vector_data (List[float]): The vector associated with the Document.
            document (Document): The original document.
        """
        super().__init__(
            page_content="", vector_data=vector_data, document=document, **kwargs
        )

    @property
    def vector_dim(self):
        return len(self.vector_data)

    def __str__(self) -> str:
        return self.document.__str__() + f" (VECTORIZED dim={self.vector_dim})"



from typing import Any, List, Dict, Tuple, Literal, Optional
from langchain_core.documents import Document


class OSEntityDocument(Document):
    type: Literal["OSEntityDocument"] = "OSEntityDocument"
    record: Dict[str, Any]
    relationships: List[Tuple[Dict[str, Any], Dict[str, Any]]]
    annotations: Dict[str, Any]
    attachment: Optional[bytes] = None

    def __init__(
        self,
        record: Dict[str, Any],
        relationships: List[Tuple[Dict[str, Any], Dict[str, Any]]],
        annotations: Dict[str, Any],
        attachment: Optional[bytes] = None,
        **kwargs: Any,
    ):
        """
        A document class designed for an Octostar entity.

        Args:
            record (Dict[str, Any]): The entity record.
            relationships (List[Tuple[Dict[str, Any], Dict[str, Any]]]): A list of entity relationships, each expressed
                as a tuple (relationship_record, target_concept_record). Either of these two elements can be None.
            annotations (Dict[str, Any]): The entity's pipeline annotations.
            attachment: Optional[bytes]: The entity's attachment, if any.
        """
        super().__init__(
            record=record,
            relationships=relationships,
            annotations=annotations,
            attachment=attachment,
            page_content="",
            **kwargs,
        )

    def __str__(self) -> str:
        return (
            f"entity_type='{self.record['os_concept']}', "
            f"n_fields='{len([v for v in self.record.values() if v is not None])}"
            f"n_annotations={len([v for v in self.annotations.values() if v is not None])}"
            f"attachment_size={len(self.attachment) if self.attachment else 0}"
        )



from langchain_openai import OpenAIEmbeddings as _OpenAIEmbeddings
from typing import Any, ClassVar


try:
    from langchain_huggingface import HuggingFaceEmbeddings as _HuggingFaceEmbeddings

    class HuggingFaceEmbeddings(_HuggingFaceEmbeddings):
        model_version: str

        def __init__(self, **kwargs: Any):
            """Initialize the sentence_transformer."""
            kwargs.setdefault("model_version", "1")
            super().__init__(**kwargs)

        def get_embedding_dimension(self) -> int:
            return self._client.get_sentence_embedding_dimension()

except ImportError:

    class MissingHuggingFaceEmbeddings:
        def __init__(self, *args, **kwargs):
            raise ImportError(
                "HuggingFaceEmbeddings requires 'langchain_huggingface' to be installed. "
                "Install it using: pip install octostar-ai[gpu]"
            )

    HuggingFaceEmbeddings = MissingHuggingFaceEmbeddings


class OpenAIEmbeddings(_OpenAIEmbeddings):
    # Define static variables as class variables
    DEFAULT_MODEL: ClassVar[str] = "text-embedding-3-small"
    DEFAULT_DIMENSIONS: ClassVar[int] = 1536

    model_version: str
    model_name: str

    def __init__(self, **kwargs: Any):
        if not kwargs:
            kwargs = {}
        kwargs.setdefault("model_version", "1")
        kwargs.setdefault("model", self.DEFAULT_MODEL)
        kwargs.setdefault("model_name", self.DEFAULT_MODEL)
        kwargs.setdefault("dimensions", self.DEFAULT_DIMENSIONS)

        super().__init__(**kwargs)

    def get_embedding_dimension(self) -> int:
        return self.dimensions



from pydantic import BaseModel, Field, ConfigDict
from typing import List, Dict, Any
import re


class VectorDict(BaseModel):
    """Generic class representing a vector."""

    class Metadata(BaseModel):
        model_config = ConfigDict(extra="allow")

        model_name: str
        model_version: str
        dim: int

    type: str = Field(alias="#type")
    data: List[float]
    metadata: Metadata

    @staticmethod
    def safe_string(string: str) -> str:
        """Converts a string so it only contains identifiers-like characters."""
        string = re.sub("[^0-9a-zA-Z_\-\.\\\/]+", "", string)
        string = re.sub("[#_\-\.\\\/]+", "_", string)
        if len(string) > 100:
            string = string[:100]
        return string

    def model_post_init(self, __context):
        assert self.type == "VECTOR"
        assert len(self.data) == self.metadata.dim and self.metadata.dim > 0
        assert self.metadata.model_name and self.metadata.model_version
        self.metadata.model_name = VectorDict.safe_string(self.metadata.model_name)
        self.metadata.model_version = VectorDict.safe_string(
            self.metadata.model_version
        )


def validate_and_format_vector(source: dict) -> dict:
    """Validates that an input dictionary is a valid vector, and converts its metadata to be safe strings."""
    return VectorDict(**source).model_dump(by_alias=True)

def validate_and_format_vector_meta(source: dict) -> dict:
    """Validates that an input dictionary is a valid vector metadata, and converts it to be safe strings."""
    # Create a mock vector
    vector = {
        '#type': 'VECTOR',
        'metadata': source,
        'data': [0]*int(source.get('dim', 1))
    }
    vector = VectorDict(**vector).model_dump(by_alias=True)
    return vector['metadata']



from langchain_text_splitters.json import RecursiveJsonSplitter
from langchain_text_splitters import RecursiveCharacterTextSplitter
from langchain_core.documents import Document
import re
from collections import Counter
from ..documents.commons import JsonDocument, TextDocument, Document
from ..documents.entities import OSEntityDocument
from .commons import MapSplitter, JsonSplitter, ChainedSplitter
from .entities import OSEntitySplitter
from .postprocessors.commons import ReEnumeratePostProcessor


def default_entity_splitter():
    """
    Get the default entity splitter for an entity (OSEntityDocument).
    The splitter takes care of chunking the entity record, metadata, relationships, and so on.
    """

    def _none_map(doc: Document, **kwargs) -> TextDocument:
        return TextDocument(page_content="", metadata=doc.metadata)

    def _tags_map(doc: JsonDocument, **kwargs) -> TextDocument:
        json_content = doc.json_content
        tags = []
        for elem in json_content["content"]:
            if elem.get("label"):
                tags.append(str(elem.get("label")))
        tags = Counter(tags)
        tags_text = "Image tags: " + ", ".join(
            f"{item} ({count})" for item, count in tags.items()
        )
        return TextDocument(page_content=tags_text, metadata=doc.metadata)

    # Chunkers for specific fields. For linkcharts, we filter out unnecessary fields
    linkchart_keep_base_fields = [
        r"nodes\[\d+\]._entityData",
        r"edges\[\d+\]._edgeData",
    ]
    linkchart_keep_relationship_fields = [
        r"relationship_name",
        r"inverse_name",
        r"description",
    ]
    special_splitters = {
        "annotations.extract:txt": (
            TextDocument,
            ChainedSplitter(
                RecursiveCharacterTextSplitter(separators=["\n\n", "\n", " ", ""]),
                ReEnumeratePostProcessor("chunk_id"),
            ),
        ),
        "annotations.image:annotations": (JsonDocument, MapSplitter(_tags_map)),
        "annotations.extract:metadata": (
            JsonDocument,
            OSEntitySplitter.get_splitter_json_to_text(),
        ),
        "record.os_item_content": (JsonDocument, None),
        "record.os_item_content.data": (
            JsonDocument,
            OSEntitySplitter.get_splitter_json_to_text(
                ignore_fields=[
                    re.compile(r".*#relationships_fetched"),
                    re.compile(
                        r"edges\[\d+\]._edgeData.relationship.(?!("
                        + "|".join(linkchart_keep_relationship_fields)
                        + r")).*"
                    ),
                    re.compile(
                        r"^(?!(" + "|".join(linkchart_keep_base_fields) + r")).*"
                    ),
                ]
            ),
        ),
        "record.metadata": (JsonDocument, OSEntitySplitter.get_splitter_json_to_text()),
    }

    # Chunker for the first chunk (no-op)
    base_splitters = MapSplitter(_none_map)

    # Chunkers for the record fields. Notice we ignore the fields treated specially above
    record_splitters = ChainedSplitter(
        OSEntitySplitter.get_splitter_json_to_text(
            ignore_fields=[
                k[7:] for k in special_splitters.keys() if k.startswith("record.")
            ]
        ),
        ReEnumeratePostProcessor("chunk_id"),
    )

    # Chunkers for the entity's relationships. They are also json objects, so it's more of the same. We only keep
    # the relationship label and a few other fields.
    relationship_keep_fields = [
        r"relationship\.(entity_label|entity_type|os_workspace|os_entity_uid|os_entity_uid_from|os_entity_uid_to|os_relationship_name)",
        r"other_entity\.(entity_label|entity_type|os_workspace|os_entity_uid)",
        r"direction",
    ]
    relationship_splitters = ChainedSplitter(
        MapSplitter(OSEntitySplitter.parse_relationship_as_text),
        OSEntitySplitter.get_splitter_json_to_text(
            ignore_fields=[
                re.compile(r"^(?!(" + "|".join(relationship_keep_fields) + r")).*")
            ]
        ),
        ## TODO: Add a recombinator to merge relationships from the same workspace
        ReEnumeratePostProcessor("chunk_id", "os_workspace"),
    )

    # Chunkers for the entity's attachment (if any). For now we don't use it.
    attachment_splitters = None

    # Chunkers for the entity's AI annotations. We don't use it since we'd rather use the special splitters for each field.
    annotations_splitter = None

    return OSEntitySplitter(
        base_splitters,
        record_splitters,
        relationship_splitters,
        attachment_splitters,
        annotations_splitter,
        special_splitters,
    )


def default_json_splitter():
    """
    Get the default json splitter for a JSON-compatible input (JSONDocument).
    """
    return ChainedSplitter(
        JsonSplitter(RecursiveJsonSplitter(), as_text=True),
        RecursiveCharacterTextSplitter(separators=["\n\n", "\n", " ", ""]),
    )


def default_text_splitter():
    """
    Get the default text splitter for a string input (TextDocument).
    """
    return RecursiveCharacterTextSplitter(separators=["\n\n", "\n", " ", ""])



from typing import Optional
from octostar_ai.utils.gpu import get_device


def default_rag_embedder(openai_api_key: Optional[str] = None):
    """Get the default embedder for Retrieval-Augmented Generation (RAG)."""
    if openai_api_key:
        from octostar_ai.runnables.embedders.commons import OpenAIEmbeddings

        return OpenAIEmbeddings(
            model="text-embedding-3-small", model_version="1", api_key=openai_api_key
        )

    from octostar_ai.runnables.embedders.commons import HuggingFaceEmbeddings

    model_name = "sentence-transformers/all-mpnet-base-v2"
    model_kwargs = {"device": get_device()}
    encode_kwargs = {"normalize_embeddings": False}
    return HuggingFaceEmbeddings(
        model_name=model_name, model_kwargs=model_kwargs, encode_kwargs=encode_kwargs
    )



from typing import Optional, Dict, Tuple, Type, Sequence, Any
from warnings import warn
from langchain_core.documents import Document
from ..documents.commons import JsonDocument, TextDocument
from ..documents.entities import OSEntityDocument
from .postprocessors.commons import TrimEmptyPostProcessor, ReEnumeratePostProcessor
from .commons import Splitter, JsonSplitter, ChainedSplitter
from langchain_text_splitters.json import RecursiveJsonSplitter
from langchain_text_splitters import RecursiveCharacterTextSplitter
import re
from typing import List, Union


class OSEntitySplitter(Splitter):
    """Interface for splitting Octostar entities into chunks."""

    def __init__(
        self,
        base_splitter: Splitter,
        record_splitter: Optional[Splitter],
        relationship_splitter: Optional[Splitter],
        attachment_splitter: Optional[Splitter],
        annotations_splitter: Optional[Splitter],
        special_splitters: Optional[
            Dict[
                str,
                Tuple[
                    Type[Document],
                    Splitter,
                ],
            ]
        ],
    ):
        """Create a new OSEntitySplitter.

        Args:
            base_splitter: A JSON splitter which should produce the most important chunk to identify the entity
            record_splitter: A JSON splitter which will be used for the entity record
            relationship_splitter: A JSON splitter which will be used for the entity relationships records
            attachment_splitter: A binary splitter to convert binary contents into a sequence of Document
            annotations_splitter: A JSON splitter which will be used for the entity's pipeline annotations
            special_splitters: A mapping from field names to splitters, for record fields that should be handled differently.
                Nested fields can be accessed with the "." notation.
                Preface the field name with either "record", "relationships", "attachment" or "annotations".
        """
        self.base_splitter = base_splitter
        self.record_splitter = record_splitter
        self.relationship_splitter = relationship_splitter
        self.attachment_splitter = attachment_splitter
        self.annotations_splitter = annotations_splitter
        self.special_splitters = special_splitters or {}

    def _transform_entity(self, entity: OSEntityDocument):
        all_chunks = []
        assert self.base_splitter
        base_docs = [JsonDocument(entity.record, metadata={"type": "base"})]
        base_docs = self.base_splitter.transform_documents(base_docs, entity=entity)
        if len(base_docs) > 1:
            warn(f"Base chunk is of length {len(base_docs)}, expected 1!")
        assert bool(base_docs)
        all_chunks.append(base_docs[0])
        if self.record_splitter:
            record_docs = [JsonDocument(entity.record, metadata={"type": "record"})]
            record_docs = self.record_splitter.transform_documents(
                record_docs, entity=entity
            )
            all_chunks.extend(record_docs)
        if self.annotations_splitter:
            annotations_docs = [
                JsonDocument(entity.annotations, metadata={"type": "annotation"})
            ]
            annotations_docs = self.annotations_splitter.transform_documents(
                record_docs, entity=entity
            )
            all_chunks.extend(annotations_docs)
        if self.relationship_splitter:
            rel_docs = [
                JsonDocument(
                    {"relationship": rel[0], "other_entity": rel[1]},
                    metadata={
                        "type": "relationship",
                        "os_workspace": rel[1].get("os_worspace"),
                        "os_relationship_workspace": rel[0].get("os_workspace"),
                    },
                )
                for rel in entity.relationships
            ]
            rel_docs = self.relationship_splitter.transform_documents(
                rel_docs, entity=entity
            )
            all_chunks.extend(rel_docs)
        all_special_docs = []
        for field_name, (doc_cls, splitter) in self.special_splitters.items():
            if splitter:
                field_base_key, field_name = field_name.split(".", 1)
                field = getattr(entity, field_base_key)
                is_field_valid = True
                for field_key in field_name.split("."):
                    try:
                        if isinstance(field, dict):
                            field = field[field_key]
                        elif isinstance(field, list):
                            field = field[int(field_key)]
                        else:
                            raise KeyError(field_key)
                    except KeyError:
                        is_field_valid = False
                        break
                if not is_field_valid:
                    warn(f'"{field_name}" not found in entity {field_base_key}!')
                    continue
                special_docs = [
                    doc_cls(
                        field,
                        metadata={
                            "type": "other",
                            "field_name": field_base_key + "." + field_name,
                        },
                    )
                ]
                special_docs = splitter.transform_documents(special_docs, entity=entity)
                all_special_docs.extend(special_docs)
        all_chunks.extend(all_special_docs)
        all_chunks = TrimEmptyPostProcessor().transform_documents(all_chunks)
        return all_chunks

    def transform_documents(
        self, documents: Sequence[OSEntityDocument], **kwargs: Any
    ) -> Sequence[TextDocument]:
        """Transform a list of documents.

        Args:
            documents: A sequence of Octostar entities to be transformed.

        Returns:
            A sequence of flat Documents (text + metadata).
        """
        entities = documents
        all_chunks = []
        for entity in entities:
            all_chunks.extend(self._transform_entity(entity))
        return all_chunks

    @staticmethod
    def prepend_base_entity_info(doc: Document, entity: OSEntityDocument):
        ## TODO: add some more fields for files
        base_doc = f"{entity.record['entity_label']} ({entity.record['entity_type']}) contained in Workspace {entity.record['os_workspace']}"
        doc.page_content = base_doc + "\n" + doc.page_content
        doc.page_content = doc.page_content.strip()
        return doc

    @staticmethod
    def get_splitter_json_to_text(
        ignore_fields: List[Union[str, re.Pattern]] = [],
        separators: List[str] = ["\n\n", "\n", " ", ""],
        re_enumerate_on: str = "chunk_id",
    ):
        return ChainedSplitter(
            JsonSplitter(
                RecursiveJsonSplitter(), as_text=True, ignore_fields=ignore_fields
            ),
            RecursiveCharacterTextSplitter(separators=separators),
            ReEnumeratePostProcessor(index_name=re_enumerate_on),
        )

    @staticmethod
    def parse_relationship_as_text(rel: JsonDocument, entity: OSEntityDocument):
        contents = rel.json_content
        direction = None
        if (
            contents["relationship"]["os_entity_uid_from"]
            == contents["relationship"]["os_entity_uid_from"]
        ):
            direction = 0
        elif (
            contents["relationship"]["os_entity_uid_to"]
            == contents["other_entity"]["os_entity_uid"]
        ):
            direction = 1
        elif (
            contents["relationship"]["os_entity_uid_from"]
            == contents["other_entity"]["os_entity_uid"]
        ):
            direction = -1
        if not direction:
            direction = 0
        this_entity_txt = (
            f"{entity.record['entity_label']} ({entity.record['os_entity_uid']})"
        )
        other_entity_txt = f"{contents['other_entity']['entity_label']} ({contents['other_entity']['os_entity_uid']})"
        relationship_txt = contents["relationship"]["os_relationship_name"].replace(
            "_", " "
        )
        text = ""
        match direction:
            case 1:
                text = f"{this_entity_txt} {relationship_txt} {other_entity_txt}"
            case -1:
                text = f"{other_entity_txt} {relationship_txt} {this_entity_txt}"
            case 0:
                text = f"{relationship_txt} between {this_entity_txt} and {other_entity_txt}"
        contents["direction"] = text
        return rel



def get_device() -> str:
    try:
        import torch
    except ImportError:
        return "cpu"

    if torch.cuda.is_available():
        device = f"cuda:{torch.cuda.current_device()}"
    elif torch.backends.mps.is_available():
        device = "mps"
    else:
        device = "cpu"
    return device



import copy
from langchain_core.documents.transformers import BaseDocumentTransformer
from langchain_text_splitters.json import RecursiveJsonSplitter
from collections.abc import Sequence
from typing import Any, List, Union, Callable, Concatenate
from langchain_core.documents import Document
import re
import yaml

from ..documents.commons import JsonDocument, TextDocument


class Splitter(BaseDocumentTransformer):
    """Unified interface for splitting data into chunks."""

    def transform_documents(
        self, documents: Sequence[Document], **kwargs: Any
    ) -> Sequence[Document]:
        raise NotImplementedError()


class MapSplitter(Splitter):
    """A splitter which applies a function to each document. Does not change the number of documents or their order."""

    def __init__(
        self, map_fn: Callable[Concatenate[Document, ...], Document], **map_kwargs
    ):
        """
        Args:
            map_fn (Callable[Concatenate[Document, ...], Document]): A callable applying a transform on a document.
                The callable must also take in arbitrary **kwargs.
            map_kwargs: Static kwargs to pass to the mapping callable. Further arguments can be specified dynamically
                when invoking self.transform_documents().
        """
        self.map_fn = map_fn
        self.map_kwargs = map_kwargs

    def transform_documents(
        self, documents: Sequence[Document], **kwargs: Any
    ) -> Sequence[Document]:
        return [self.map_fn(doc, **{**self.map_kwargs, **kwargs}) for doc in documents]


class ChainedSplitter(Splitter):
    """A splitter which applies multiple splitters sequentially, one after the other."""

    def __init__(self, *splitters: List[Splitter]):
        """
        Args:
            splitters (List[Splitter]): The list of splitters to apply, in order.
        """
        self.splitters = splitters

    def transform_documents(
        self, documents: Sequence[Document], **kwargs: Any
    ) -> Sequence[Document]:
        for splitter in self.splitters:
            documents = splitter.transform_documents(documents, **kwargs)
        return documents


class MultiPassSplitter(Splitter):
    """A splitter which applies multiple splitters in parallel, as separate passes.
    The resulting set of documents is given by the concatenation of the original documents, where each
    has been passed through a different splitter."""

    def __init__(self, *splitters: List[Splitter]):
        """
        Args:
            splitters (List[Splitter]): The list of splitters to apply as separate passes.
        """
        self.splitters = splitters

    def transform_documents(
        self, documents: Sequence[Document], **kwargs: Any
    ) -> Sequence[Document]:
        all_documents = []
        for splitter in self.splitters:
            all_documents.extend(
                splitter.transform_documents(copy.deepcopy(documents), **kwargs)
            )
        return all_documents


class JsonSplitter(Splitter):
    """A splitter for JSON data. Wrapper around LangChain with more customizability."""

    def __init__(
        self,
        lc_json_splitter: RecursiveJsonSplitter,
        ignore_nulls: bool = True,
        as_text: bool = False,
        ignore_fields: List[Union[str, re.Pattern]] = [],
    ):
        """
        Args:
            splitter (RecursiveJsonSplitter): The underlying LangChain splitter to use.
            ignore_nulls (bool): If True, null values are filtered out from the input documents.
            as_text (bool): if True, the output documents are formatted as newline-indented texts, rather than
                as raw json text.
            ignore_fields (List[Union[str, re.Pattern]]): A list of field names or regex expressions which will
                be filtered out from the input documents. Use the "." notation to refer to nested fields, e.g.
                    "myfield.subfield".
        """
        self.splitter = lc_json_splitter
        self.ignore_nulls = ignore_nulls
        self.as_text = as_text
        self.ignore_fields = ignore_fields

    def transform_documents(
        self, documents: Sequence[JsonDocument], **kwargs: Any
    ) -> Sequence[Union[TextDocument, JsonDocument]]:
        all_docs = []
        for doc in documents:
            json_content = doc.json_content
            keep_fields = JsonSplitter.filter_fields(json_content, self.ignore_fields)
            json_content = JsonSplitter.filter_dict_by_keys(json_content, keep_fields)
            if self.ignore_nulls:
                json_content = JsonSplitter.filter_nulls(json_content)
            split_json = self.splitter.split_json(json_content)
            split_metadatas = [
                {**copy.deepcopy(doc.metadata), "chunk_id": i}
                for i in range(len(split_json))
            ]
            split_docs = []
            if not self.as_text:
                split_docs = [
                    JsonDocument(js, metadata=split_metadatas[i])
                    for i, js in enumerate(split_json)
                ]
            else:
                split_docs = [
                    TextDocument(
                        JsonSplitter.json_as_text(js), metadata=split_metadatas[i]
                    )
                    for i, js in enumerate(split_json)
                ]
            all_docs.extend(split_docs)
        return all_docs

    @staticmethod
    def json_as_text(d):
        return yaml.dump(d, default_flow_style=False)

    @staticmethod
    def filter_fields(d, ignores):
        flattened_fields = JsonSplitter.flatten_dict_keys(d)
        filtered_fields = []
        for item in flattened_fields:
            to_ignore = False
            for pattern in ignores:
                if (isinstance(pattern, str) and item.startswith(pattern)) or (
                    isinstance(pattern, re.Pattern) and re.match(pattern, item)
                ):
                    to_ignore = True
                    break
            if not to_ignore:
                filtered_fields.append(item)
        return filtered_fields

    @staticmethod
    def flatten_dict_keys(d, parent_key=""):
        items = []
        if isinstance(d, dict):
            for k, v in d.items():
                new_key = f"{parent_key}.{k}" if parent_key else k
                items.extend(JsonSplitter.flatten_dict_keys(v, new_key))
        elif isinstance(d, list):
            for i, v in enumerate(d):
                new_key = f"{parent_key}[{i}]"
                items.extend(JsonSplitter.flatten_dict_keys(v, new_key))
        else:
            items.append(parent_key)
        return items

    @staticmethod
    def filter_nulls(d):
        res = None
        if isinstance(d, dict):
            filtered_dict = {
                k: JsonSplitter.filter_nulls(v)
                for k, v in d.items()
                if v is not None and v != {} and v != []
            }
            res = filtered_dict if filtered_dict else None
        elif isinstance(d, list):
            filtered_list = [
                JsonSplitter.filter_nulls(v)
                for v in d
                if v is not None and v != {} and v != []
            ]
            res = filtered_list if filtered_list else None
        else:
            res = d
        if res is None:
            res = {}
        return res

    @staticmethod
    def filter_dict_by_keys(original_dict, filtered_keys):
        def _set_nested_value(d, keys, value):
            for i, key in enumerate(keys[:-1]):
                next_key = keys[i + 1]
                if isinstance(key, int):
                    if not isinstance(d, list):
                        raise TypeError(f"Expected list at {keys}, got {type(d)}")
                    while len(d) < key:
                        d.append(None)
                    if len(d) == key:
                        d.append([]) if isinstance(next_key, int) else d.append({})
                else:
                    if not isinstance(d, dict):
                        raise TypeError(f"Expected dict at {keys}, got {type(d)}")
                    if key not in d:
                        d[key] = [] if isinstance(next_key, int) else {}
                d = d[key]
            last_key = keys[-1]
            if isinstance(last_key, int):
                if not isinstance(d, list):
                    raise TypeError(f"Expected list at {keys}, got {type(d)}")
                if len(d) <= last_key:
                    d.extend([None] * (last_key - len(d) + 1))
                d[last_key] = value
            else:
                if not isinstance(d, dict):
                    raise TypeError(f"Expected dict at {keys}, got {type(d)}")
                d[last_key] = value

        def _parse_key(key):
            """
            Parse a string key like 'a[0].b[1]' into a list of parts: ['a', 0, 'b', 1]
            """
            parts = []
            for part in re.finditer(r"([^.[\]]+)|\[(\d+)\]", key):
                if part.group(1):  # dictionary key
                    parts.append(part.group(1))
                elif part.group(2):  # list index
                    parts.append(int(part.group(2)))
            return parts

        def _extract_nested_value(d, keys):
            """
            Extract value from the original dictionary using parsed keys.
            """
            for key in keys:
                if isinstance(key, int):  # Key is an index for a list
                    if not isinstance(d, list):
                        raise TypeError(f"Expected list at {key}, got {type(d)}")
                    d = d[key]
                else:  # Key is a dictionary key
                    if not isinstance(d, dict):
                        raise TypeError(f"Expected dict at {key}, got {type(d)}")
                    d = d[key]
            return d

        filtered_dict = {}
        for key in filtered_keys:
            parsed_keys = _parse_key(key)
            try:
                value = _extract_nested_value(original_dict, parsed_keys)
                _set_nested_value(filtered_dict, parsed_keys, value)
            except (KeyError, IndexError, TypeError):
                pass
        return filtered_dict
